import argparse
import os
import logging

import coloredlogs

from unifi.cams.hikvision import HikvisionCam
from unifi.cams.rtsp import RTSPCam
from unifi.core import Core

CAMS = {"hikvision": HikvisionCam, "rtsp": RTSPCam}

logging.basicConfig()
coloredlogs.install(level="DEBUG")


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--host", "-H", required=True, help="NVR ip address and port")
    parser.add_argument(
        "--cert",
        "-c",
        required=True,
        default="client.pem",
        help="Client certificate path",
    )
    parser.add_argument("--token", "-t", required=True, help="Adoption token")
    parser.add_argument("--mac", "-m", default="44D9E7407670", help="MAC address")
    parser.add_argument(
        "--ip", "-i", default="192.168.1.10", help="IP address of camera"
    )
    parser.add_argument(
        "--name", "-n", default="unifi-cam-proxy", help="Name of camera"
    )
    parser.add_argument(
        "--verbose", "-v", action="store_true", help="increase output verbosity"
    )

    sp = parser.add_subparsers(help="Camera implementations", dest="impl")
    for (name, impl) in CAMS.items():
        subparser = sp.add_parser(name)
        impl.add_parser(subparser)
    return parser.parse_args()


def main():
    args = parse_args()
    klass = CAMS[args.impl]

    core_logger = logging.getLogger("Core")
    logger = logging.getLogger(klass.__name__)
    if args.verbose:
        logger.setLevel(logging.DEBUG)
        core_logger.setLevel(logging.DEBUG)

    cam = klass(args, logger)
    c = Core(args, cam, core_logger)
    c.run()


if __name__ == "__main__":
    main()



""""
Helper program to inject absolute wall clock time into FLV stream for recordings
"""

import time
import struct
import sys
from flvlib import astypes
from flvlib import tags
from flvlib import primitives


def read_bytes(source, num_bytes):
    read_bytes = 0
    buf = b""
    while read_bytes < num_bytes:
        d_in = source.read(num_bytes - read_bytes)
        if d_in:
            read_bytes += len(d_in)
            buf += d_in
        else:
            return buf
    return buf


def write(data):
    sys.stdout.write(data)


def main():
    PY3K = sys.version_info >= (3, 0)

    if PY3K:
        source = sys.stdin.buffer
    else:
        if sys.platform == "win32":
            import os, msvcrt

            msvcrt.setmode(sys.stdin.fileno(), os.O_BINARY)
        source = sys.stdin

    header = read_bytes(source, 3)

    if header != "FLV":
        print("Not a valid FLV file")
        return
    write(header)

    # Skip rest of FLV header
    write(read_bytes(source, 6))

    i = 0
    while True:
        header = read_bytes(source, 15)
        if len(header) != 15:
            write(header)
            return

        # Packet structure from Wikipedia:
        #
        # Size of previous packet	uint32_be	0	For first packet set to NULL
        # Packet Type	uint8	18	For first packet set to AMF Metadata
        # Payload Size	uint24_be	varies	Size of packet data only
        # Timestamp Lower	uint24_be	0	For first packet set to NULL
        # Timestamp Upper	uint8	0	Extension to create a uint32_be value
        # Stream ID	uint24_be	0	For first stream of same type set to NULL
        # Payload Data	freeform	varies	Data as defined by packet type

        # Get payload size to know how many bytes to read
        high, low = struct.unpack(">BH", header[5:8])
        payload_size = (high << 16) + low

        # Get timestamp to inject into clock sync tag
        low_high = header[8:12]
        combined = low_high[3] + low_high[:3]
        timestamp = struct.unpack(">i", combined)[0]

        if i % 3:
            data = astypes.FLVObject()
            data["streamClock"] = int(timestamp)
            data["streamClockBase"] = 0
            data["wallClock"] = time.time() * 1000

            # Inject clock sync packet
            write(primitives.make_ui32(payload_size + 15))
            write(tags.create_script_tag("onClockSync", data, timestamp))

            # Write rest of original packet minus previous packet size
            write(header[4:])
            write(read_bytes(source, payload_size))
        else:
            # Write normal packet
            write(header)
            write(read_bytes(source, payload_size))

        i += 1


if __name__ == "__main__":
    main()



import logging
import os
import subprocess
import sys
import shutil

import tempfile
import requests
import xmltodict
from requests.auth import HTTPDigestAuth
from hikvisionapi import Client

from unifi.cams.base import UnifiCamBase

FNULL = open(os.devnull, "w")


class HikvisionCam(UnifiCamBase):
    @classmethod
    def add_parser(self, parser):
        parser.add_argument("--username", "-u", required=True, help="Camera username")
        parser.add_argument("--password", "-p", required=True, help="Camera password")

    def __init__(self, args, logger=None):
        self.logger = logger
        self.args = args
        self.dir = tempfile.mkdtemp()
        self.streams = {}
        self.cam = Client(
            "http://{}".format(self.args.ip), self.args.username, self.args.password
        )

    def get_snapshot(self):
        img_file = "{}/screen.jpg".format(self.dir)
        resp = self.cam.Streaming.channels[102].picture(
            method="get", type="opaque_data"
        )
        with open(img_file, "wb") as f:
            for chunk in resp.iter_content(chunk_size=1024):
                if chunk:
                    f.write(chunk)
        return img_file

    def get_video_settings(self):
        r = self.cam.PTZCtrl.channels[1].status(method="get")["PTZStatus"][
            "AbsoluteHigh"
        ]
        return {
            # Tilt/elevation
            "brightness": int(100 * int(r["azimuth"]) / 3600),
            # Pan/azimuth
            "contrast": int(100 * int(r["azimuth"]) / 3600),
            # Zoom
            "hue": int(100 * int(r["absoluteZoom"]) / 40),
        }

    def change_video_settings(self, options):
        tilt = int((900 * int(options["brightness"])) / 100)
        pan = int((3600 * int(options["contrast"])) / 100)
        zoom = int((40 * int(options["hue"])) / 100)

        self.logger.info("Moving to %s:%s:%s", pan, tilt, zoom)
        req = {
            "PTZData": {
                "@version": "2.0",
                "@xmlns": "http://www.hikvision.com/ver20/XMLSchema",
                "AbsoluteHigh": {
                    "absoluteZoom": str(zoom),
                    "azimuth": str(pan),
                    "elevation": str(tilt),
                },
            }
        }
        self.cam.PTZCtrl.channels[1].absolute(
            method="put", data=xmltodict.unparse(req, pretty=True)
        )

    def start_video_stream(self, stream_name, video_mode):
        channel = 1
        if video_mode == "video3":
            channel = 2

        vid_src = "rtsp://{}:{}@{}:554/Streaming/Channels/{}/".format(
            self.args.username, self.args.password, self.args.ip, channel
        )

        cmd = 'ffmpeg -y -f lavfi -i aevalsrc=0 -i "{}" -vcodec copy -use_wallclock_as_timestamps 1 -strict -2 -c:a aac -metadata streamname={} -f flv - | {} -m unifi.clock_sync | nc {} 6666'.format(
            vid_src, stream_name, sys.executable, self.args.host
        )
        self.logger.info("Spawning ffmpeg: %s", cmd)
        if (
            stream_name not in self.streams
            or self.streams[stream_name].poll() is not None
        ):
            self.streams[stream_name] = subprocess.Popen(
                cmd, stdout=FNULL, stderr=subprocess.STDOUT, shell=True
            )



# import asyncio
import json
import ssl
import time
import requests
import os
import threading
import websocket

OPCODE_DATA = (websocket.ABNF.OPCODE_TEXT, websocket.ABNF.OPCODE_BINARY)


class Core(object):
    def __init__(self, args, camera, logger):
        self.host = args.host
        self.cert = args.cert
        self.token = args.token
        self.mac = args.mac
        self.name = args.name
        self.cam_ip = args.ip
        self.cam = camera
        self.logger = logger
        self._msg_id = 0
        self.init_time = time.time()
        self.pulse_interval = 0
        self.streams = {}
        self.version = "UVC.S2L.v4.14.14.67.037e886.190630.1017"

    def gen_msg_id(self):
        self._msg_id += 1
        return self._msg_id

    def recv(self, ws):
        try:
            frame = ws.recv_frame()
        except websocket.WebSocketException:
            return websocket.ABNF.OPCODE_CLOSE, None
        if not frame:
            raise websocket.WebSocketException("Not a valid frame %s" % frame)
        elif frame.opcode in OPCODE_DATA:
            return frame.opcode, frame.data
        elif frame.opcode == websocket.ABNF.OPCODE_CLOSE:
            ws.send_close()
            return frame.opcode, None
        elif frame.opcode == websocket.ABNF.OPCODE_PING:
            ws.pong(frame.data)
            return frame.opcode, frame.data

        return frame.opcode, frame.data

    def init_adoption(self, ws):
        self.logger.info(
            "Initiating adoption with token [%s] and mac [%s]", self.token, self.mac
        )
        self.send(
            ws,
            {
                "from": "ubnt_avclient",
                "to": "UniFiVideo",
                "responseExpected": False,
                "functionName": "ubnt_avclient_hello",
                "payload": {
                    "adoptionCode": self.token,
                    "connectionHost": self.host,
                    "connectionSecurePort": 7442,
                    "fwVersion": self.version,
                    "hwrev": 19,
                    "idleTime": 191.96,
                    "ip": self.cam_ip,
                    "mac": self.mac,
                    "model": "UVC G3",
                    "name": self.name,
                    "protocolVersion": 67,
                    "rebootTimeoutSec": 30,
                    "semver": "v4.4.8",
                    "totalLoad": 0.5474,
                    "upgradeTimeoutSec": 150,
                    "uptime": self.get_uptime(),
                },
                "messageId": self.gen_msg_id(),
                "inResponseTo": 0,
            },
        )

    def process_param_agreement(self, msg):
        return {
            "from": "ubnt_avclient",
            "functionName": "ubnt_avclient_paramAgreement",
            "inResponseTo": msg["messageId"],
            "messageId": self.gen_msg_id(),
            "payload": {"authToken": self.token},
            "responseExpected": False,
            "to": "UniFiVideo",
        }

    def process_upgrade(self, msg):
        url = msg["payload"]["uri"]
        headers = {"Range": "bytes=0-100"}
        r = requests.get(url, headers=headers, verify=False)

        # Parse the new version string from the upgrade binary
        version = ""
        for i in range(0, 50):
            b = r.content[4 + i]
            if b != b"\x00":
                version += b
        self.version = version
        return

    def process_isp_settings(self, msg):
        payload = {
            "aeMode": "auto",
            "aeTargetPercent": 50,
            "aggressiveAntiFlicker": 0,
            "brightness": 50,
            "contrast": 50,
            "criticalTmpOfProtect": 40,
            "darkAreaCompensateLevel": 0,
            "denoise": 50,
            "enable3dnr": 1,
            "enableMicroTmpProtect": 1,
            "enablePauseMotion": 0,
            "flip": 0,
            "focusMode": "ztrig",
            "focusPosition": 0,
            "forceFilterIrSwitchEvents": 0,
            "hue": 50,
            "icrLightSensorNightThd": 0,
            "icrSensitivity": 0,
            "irLedLevel": 215,
            "irLedMode": "auto",
            "irOnStsBrightness": 0,
            "irOnStsContrast": 0,
            "irOnStsDenoise": 0,
            "irOnStsHue": 0,
            "irOnStsSaturation": 0,
            "irOnStsSharpness": 0,
            "irOnStsWdr": 0,
            "irOnValBrightness": 50,
            "irOnValContrast": 50,
            "irOnValDenoise": 50,
            "irOnValHue": 50,
            "irOnValSaturation": 50,
            "irOnValSharpness": 50,
            "irOnValWdr": 1,
            "mirror": 0,
            "queryIrLedStatus": 0,
            "saturation": 50,
            "sharpness": 50,
            "touchFocusX": 1001,
            "touchFocusY": 1001,
            "wdr": 1,
            "zoomPosition": 0,
        }
        payload.update(self.cam.get_video_settings())
        return {
            "from": "ubnt_avclient",
            "to": "UniFiVideo",
            "responseExpected": False,
            "functionName": "ResetIspSettings",
            "payload": payload,
            "messageId": self.gen_msg_id(),
            "inResponseTo": msg["messageId"],
        }

    def process_video_settings(self, msg):
        # self.cam.set_video_settings(msg['payload'])
        vid_dst = {
            "video1": ["file:///dev/null"],
            "video2": ["file:///dev/null"],
            "video3": ["file:///dev/null"],
        }

        if msg["payload"] is not None:
            for k, v in msg["payload"]["video"].items():
                if v:
                    if "avSerializer" in v:
                        vid_dst[k] = v["avSerializer"]["destinations"]
                        if "parameters" in v["avSerializer"]:
                            self.streams[k] = stream = v["avSerializer"]["parameters"][
                                "streamName"
                            ]
                            self.cam.start_video_stream(stream, k)

        return {
            "from": "ubnt_avclient",
            "to": "UniFiVideo",
            "responseExpected": False,
            "functionName": "ChangeVideoSettings",
            "payload": {
                "audio": {
                    "bitRate": 32000,
                    "channels": 1,
                    "description": "audio track",
                    "enableTemporalNoiseShaping": False,
                    "enabled": True,
                    "mode": 0,
                    "quality": 0,
                    "sampleRate": 11025,
                    "type": "aac",
                    "volume": 100,
                },
                "firmwarePath": "/lib/firmware/",
                "video": {
                    "enableHrd": False,
                    "hdrMode": 0,
                    "lowDelay": False,
                    "mjpg": {
                        "avSerializer": {
                            "destinations": [
                                "file:///tmp/snap.jpeg",
                                "file:///tmp/snap_av.jpg",
                            ],
                            "parameters": {
                                "audioId": 1000,
                                "enableTimestampsOverlapAvoidance": False,
                                "suppressAudio": True,
                                "suppressVideo": False,
                                "videoId": 1001,
                            },
                            "type": "mjpg",
                        },
                        "bitRateCbrAvg": 500000,
                        "bitRateVbrMax": 500000,
                        "bitRateVbrMin": None,
                        "description": "JPEG pictures",
                        "enabled": True,
                        "fps": 5,
                        "height": 720,
                        "isCbr": False,
                        "maxFps": 5,
                        "minClientAdaptiveBitRate": 0,
                        "minMotionAdaptiveBitRate": 0,
                        "nMultiplier": None,
                        "name": "mjpg",
                        "quality": 80,
                        "sourceId": 3,
                        "streamId": 8,
                        "streamOrdinal": 3,
                        "type": "mjpg",
                        "validBitrateRangeMax": 6000000,
                        "validBitrateRangeMin": 32000,
                        "width": 1280,
                    },
                    "video1": {
                        "M": 1,
                        "N": 30,
                        "avSerializer": {
                            "destinations": vid_dst["video1"],
                            "parameters": None
                            if "video1" not in self.streams
                            else {
                                "audioId": None,
                                "streamName": self.streams["video1"],
                                "suppressAudio": None,
                                "suppressVideo": None,
                                "videoId": None,
                            },
                            "type": "extendedFlv",
                        },
                        "bitRateCbrAvg": 1400000,
                        "bitRateVbrMax": 2800000,
                        "bitRateVbrMin": 48000,
                        "description": "Hi quality video track",
                        "enabled": True,
                        "fps": 15,
                        "gopModel": 0,
                        "height": 720,
                        "horizontalFlip": False,
                        "isCbr": False,
                        "maxFps": 30,
                        "minClientAdaptiveBitRate": 0,
                        "minMotionAdaptiveBitRate": 0,
                        "nMultiplier": 6,
                        "name": "video1",
                        "sourceId": 0,
                        "streamId": 1,
                        "streamOrdinal": 0,
                        "type": "h264",
                        "validBitrateRangeMax": 2800000,
                        "validBitrateRangeMin": 32000,
                        "validFpsValues": [
                            1,
                            2,
                            3,
                            4,
                            5,
                            6,
                            8,
                            9,
                            10,
                            12,
                            15,
                            16,
                            18,
                            20,
                            24,
                            25,
                            30,
                        ],
                        "verticalFlip": False,
                        "width": 1280,
                    },
                    "video2": {
                        "M": 1,
                        "N": 30,
                        "avSerializer": {
                            "destinations": vid_dst["video2"],
                            "parameters": None
                            if "video2" not in self.streams
                            else {
                                "audioId": None,
                                "streamName": self.streams["video2"],
                                "suppressAudio": None,
                                "suppressVideo": None,
                                "videoId": None,
                            },
                            "type": "extendedFlv",
                        },
                        "bitRateCbrAvg": 500000,
                        "bitRateVbrMax": 1200000,
                        "bitRateVbrMin": 48000,
                        "currentVbrBitrate": 1200000,
                        "description": "Medium quality video track",
                        "enabled": True,
                        "fps": 15,
                        "gopModel": 0,
                        "height": 400,
                        "horizontalFlip": False,
                        "isCbr": False,
                        "maxFps": 30,
                        "minClientAdaptiveBitRate": 0,
                        "minMotionAdaptiveBitRate": 0,
                        "nMultiplier": 6,
                        "name": "video2",
                        "sourceId": 1,
                        "streamId": 2,
                        "streamOrdinal": 1,
                        "type": "h264",
                        "validBitrateRangeMax": 1500000,
                        "validBitrateRangeMin": 32000,
                        "validFpsValues": [
                            1,
                            2,
                            3,
                            4,
                            5,
                            6,
                            8,
                            9,
                            10,
                            12,
                            15,
                            16,
                            18,
                            20,
                            24,
                            25,
                            30,
                        ],
                        "verticalFlip": False,
                        "width": 720,
                    },
                    "video3": {
                        "M": 1,
                        "N": 30,
                        "avSerializer": {
                            "destinations": vid_dst["video3"],
                            "parameters": None
                            if "video3" not in self.streams
                            else {
                                "audioId": None,
                                "streamName": self.streams["video3"],
                                "suppressAudio": None,
                                "suppressVideo": None,
                                "videoId": None,
                            },
                            "type": "extendedFlv",
                        },
                        "bitRateCbrAvg": 300000,
                        "bitRateVbrMax": 200000,
                        "bitRateVbrMin": 48000,
                        "currentVbrBitrate": 200000,
                        "description": "Low quality video track",
                        "enabled": True,
                        "fps": 15,
                        "gopModel": 0,
                        "height": 360,
                        "horizontalFlip": False,
                        "isCbr": False,
                        "maxFps": 30,
                        "minClientAdaptiveBitRate": 0,
                        "minMotionAdaptiveBitRate": 0,
                        "nMultiplier": 6,
                        "name": "video3",
                        "sourceId": 2,
                        "streamId": 4,
                        "streamOrdinal": 2,
                        "type": "h264",
                        "validBitrateRangeMax": 750000,
                        "validBitrateRangeMin": 32000,
                        "validFpsValues": [
                            1,
                            2,
                            3,
                            4,
                            5,
                            6,
                            8,
                            9,
                            10,
                            12,
                            15,
                            16,
                            18,
                            20,
                            24,
                            25,
                            30,
                        ],
                        "verticalFlip": False,
                        "width": 640,
                    },
                    "vinFps": 30,
                },
            },
            "messageId": self.gen_msg_id(),
            "inResponseTo": msg["messageId"],
        }

    def process_device_settings(self, msg):
        return {
            "from": "ubnt_avclient",
            "to": "UniFiVideo",
            "responseExpected": False,
            "functionName": "ChangeDeviceSettings",
            "payload": {
                "name": self.name,
                "region": msg["payload"]["region"],
                "timezone": "PST8PDT,M3.2.0,M11.1.0",
            },
            "messageId": self.gen_msg_id(),
            "inResponseTo": msg["messageId"],
        }

    def process_osd_settings(self, msg):
        return {
            "from": "ubnt_avclient",
            "to": "UniFiVideo",
            "responseExpected": False,
            "functionName": "ChangeOsdSettings",
            "payload": {
                "_1": {
                    "enableDate": 1,
                    "enableLogo": 1,
                    "enableReportdStatsLevel": 0,
                    "enableStreamerStatsLevel": 0,
                    "tag": self.name,
                },
                "_2": {
                    "enableDate": 1,
                    "enableLogo": 1,
                    "enableReportdStatsLevel": 0,
                    "enableStreamerStatsLevel": 0,
                    "tag": self.name,
                },
                "_3": {
                    "enableDate": 1,
                    "enableLogo": 1,
                    "enableReportdStatsLevel": 0,
                    "enableStreamerStatsLevel": 0,
                    "tag": self.name,
                },
                "_4": {
                    "enableDate": 1,
                    "enableLogo": 1,
                    "enableReportdStatsLevel": 0,
                    "enableStreamerStatsLevel": 0,
                    "tag": self.name,
                },
                "enableOverlay": 1,
                "logoScale": 50,
                "overlayColorId": 0,
                "textScale": 50,
                "useCustomLogo": 0,
            },
            "messageId": self.gen_msg_id(),
            "inResponseTo": msg["messageId"],
        }

    def process_network_status(self, msg):
        return {
            "from": "ubnt_avclient",
            "to": "UniFiVideo",
            "responseExpected": False,
            "functionName": "NetworkStatus",
            "payload": {
                "connectionState": 2,
                "connectionStateDescription": "CONNECTED",
                "defaultInterface": "eth0",
                "dhcpLeasetime": 86400,
                "dnsServer": "8.8.8.8 4.2.2.2",
                "gateway": "192.168.103.1",
                "ipAddress": self.cam_ip,
                "linkDuplex": 1,
                "linkSpeedMbps": 100,
                "mode": "dhcp",
                "networkMask": "255.255.255.0",
            },
            "messageId": self.gen_msg_id(),
            "inResponseTo": msg["messageId"],
        }

    def process_sound_led_settings(self, msg):
        return {
            "from": "ubnt_avclient",
            "to": "UniFiVideo",
            "responseExpected": False,
            "functionName": "ChangeSoundLedSettings",
            "payload": {
                "ledFaceAlwaysOnWhenManaged": 1,
                "ledFaceEnabled": 1,
                "speakerEnabled": 1,
                "speakerVolume": 100,
                "systemSoundsEnabled": 1,
                "userLedBlinkPeriodMs": 0,
                "userLedColorFg": "blue",
                "userLedOnNoff": 1,
            },
            "messageId": self.gen_msg_id(),
            "inResponseTo": msg["messageId"],
        }

    def process_change_isp_settings(self, msg):
        payload = {
            "aeMode": "auto",
            "aeTargetPercent": 50,
            "aggressiveAntiFlicker": 0,
            "brightness": 50,
            "contrast": 50,
            "criticalTmpOfProtect": 40,
            "dZoomCenterX": 50,
            "dZoomCenterY": 50,
            "dZoomScale": 0,
            "dZoomStreamId": 4,
            "darkAreaCompensateLevel": 0,
            "denoise": 50,
            "enable3dnr": 1,
            "enableExternalIr": 0,
            "enableMicroTmpProtect": 1,
            "enablePauseMotion": 0,
            "flip": 0,
            "focusMode": "ztrig",
            "focusPosition": 0,
            "forceFilterIrSwitchEvents": 0,
            "hue": 50,
            "icrLightSensorNightThd": 0,
            "icrSensitivity": 0,
            "irLedLevel": 215,
            "irLedMode": "auto",
            "irOnStsBrightness": 0,
            "irOnStsContrast": 0,
            "irOnStsDenoise": 0,
            "irOnStsHue": 0,
            "irOnStsSaturation": 0,
            "irOnStsSharpness": 0,
            "irOnStsWdr": 0,
            "irOnValBrightness": 50,
            "irOnValContrast": 50,
            "irOnValDenoise": 50,
            "irOnValHue": 50,
            "irOnValSaturation": 50,
            "irOnValSharpness": 50,
            "irOnValWdr": 1,
            "lensDistortionCorrection": 1,
            "masks": None,
            "mirror": 0,
            "queryIrLedStatus": 0,
            "saturation": 50,
            "sharpness": 50,
            "touchFocusX": 1001,
            "touchFocusY": 1001,
            "wdr": 1,
            "zoomPosition": 0,
        }

        if msg["payload"]:
            self.cam.change_video_settings(msg["payload"])

        payload.update(self.cam.get_video_settings())
        return {
            "from": "ubnt_avclient",
            "to": "UniFiVideo",
            "responseExpected": False,
            "functionName": "ChangeIspSettings",
            "payload": payload,
            "messageId": self.gen_msg_id(),
            "inResponseTo": msg["messageId"],
        }

    def process_analytics_settings(self, msg):
        if msg["payload"]["sendPulse"] is 1:
            self.pulse_interval = msg["payload"]["pulsePeriodSec"]
        else:
            self.pulse_interval = 0
        return {
            "from": "ubnt_avclient",
            "to": "UniFiVideo",
            "responseExpected": False,
            "functionName": "ChangeAnalyticsSettings",
            "payload": msg["payload"],
            "messageId": self.gen_msg_id(),
            "inResponseTo": msg["messageId"],
        }

    def process_snapshot_request(self, msg):
        path = self.cam.get_snapshot()
        while not os.path.isfile(path):
            time.sleep(0.1)
        files = {"payload": (msg["payload"]["filename"], open(path, "rb"))}
        requests.post(
            msg["payload"]["uri"],
            files=files,
            data=msg["payload"]["formFields"],
            cert=self.cert,
            verify=False,
        )

    def process_time(self, msg):
        return {
            "from": "ubnt_avclient",
            "functionName": "ubnt_avclient_paramAgreement",
            "inResponseTo": msg["messageId"],
            "messageId": self.gen_msg_id(),
            "payload": {
                "monotonicMs": self.get_uptime(),
                "wallMs": int(round(time.time() * 1000)),
            },
            "responseExpected": False,
            "to": "UniFiVideo",
        }

    def process_username_password(self, msg):
        return {
            "from": "ubnt_avclient",
            "functionName": "UpdateUsernamePassword",
            "inResponseTo": msg["messageId"],
            "messageId": self.gen_msg_id(),
            "payload": {},
            "responseExpected": False,
            "to": "UniFiVideo",
        }

    def get_uptime(self):
        return time.time() - self.init_time

    def send(self, ws, msg):
        self.logger.debug("Sending: %s", msg)
        ws.send_binary(json.dumps(msg))

    def process(self, ws, msg):
        m = json.loads(msg)
        self.logger.info("Processing [%s] message", m["functionName"])
        self.logger.debug("Message contents: %s", m)

        if m["responseExpected"] == False and m["functionName"] not in [
            "GetRequest",
            "UpdateFirmwareRequest",
        ]:
            return

        res = None
        if m["functionName"] == "ubnt_avclient_hello":
            pass
        elif m["functionName"] == "ubnt_avclient_timeSync":
            pass
        elif m["functionName"] == "ubnt_avclient_time":
            res = self.process_time(m)
        elif m["functionName"] == "ubnt_avclient_paramAgreement":
            res = self.process_param_agreement(m)
        elif m["functionName"] == "ResetIspSettings":
            res = self.process_isp_settings(m)
        elif m["functionName"] == "ChangeVideoSettings":
            res = self.process_video_settings(m)
        elif m["functionName"] == "ChangeDeviceSettings":
            res = self.process_device_settings(m)
        elif m["functionName"] == "ChangeOsdSettings":
            res = self.process_osd_settings(m)
        elif m["functionName"] == "NetworkStatus":
            res = self.process_network_status(m)
        elif m["functionName"] == "ChangeSoundLedSettings":
            res = self.process_sound_led_settings(m)
        elif m["functionName"] == "ChangeIspSettings":
            res = self.process_change_isp_settings(m)
        elif m["functionName"] == "ChangeAnalyticsSettings":
            res = self.process_analytics_settings(m)
        elif m["functionName"] == "GetRequest":
            self.process_snapshot_request(m)
        elif m["functionName"] == "UpdateUsernamePassword":
            res = self.process_username_password(m)
        elif m["functionName"] == "UpdateFirmwareRequest":
            res = self.process_upgrade(m)
            return True

        if res is not None:
            self.send(ws, res)

        return False

    def send_pulse(self, ws):
        while True:
            if self.pulse_interval is not 0:
                time.sleep(self.pulse_interval)
                res = {
                    "from": "ubnt_avclient",
                    "to": "UniFiVideo",
                    "responseExpected": False,
                    "functionName": "EventAnalytics",
                    "payload": {
                        "clockBestMonotonic": 0,
                        "clockBestWall": 0,
                        "clockMonotonic": int(round(self.get_uptime())),
                        "clockWall": int(round(time.time() * 1000)),
                        "edgeType": "unknown",
                        "eventId": 9223372036854775807,
                        "eventType": "pulse",
                        "levels": {"0": 0},
                        "motionHeatmap": "",
                        "motionSnapshot": "",
                    },
                    "messageId": self.gen_msg_id(),
                    "inResponseTo": 0,
                }
                self.logger.info("Sending pulse...")
                self.send(ws, res)
            time.sleep(0.1)

    def run(self):
        uri = "wss://{}:7442/camera/1.0/ws?token={}".format(self.host, self.token)
        ssl_opts = {"cert_reqs": ssl.CERT_NONE, "certfile": self.cert}
        headers = {"camera-mac": self.mac}
        self.logger.info("Creating ws connection to %s", uri)

        while True:
            ws = websocket.create_connection(uri, sslopt=ssl_opts, header=headers)
            self.init_adoption(ws)

            self.pulse_thread = threading.Thread(target=self.send_pulse, args=(ws,))
            self.pulse_thread.daemon = True
            self.pulse_thread.start()

            while True:
                opcode, data = self.recv(ws)
                msg = None
                if opcode in OPCODE_DATA:
                    msg = data

                if msg is not None:
                    reconnect = self.process(ws, msg)
                    if reconnect:
                        self.logger.info("Reconnecting...")
                        break

                if opcode == websocket.ABNF.OPCODE_CLOSE:
                    break



import logging


class UnifiCamBase(object):
    def __init__(self, args, logger=None):
        self.args = args
        if logger is None:
            self.logger = logging.getLogger(__class__)
        else:
            self.logger = logger

    def get_snapshot(self):
        raise NotImplementedError("You need to write this!")

    def get_video_settings(self):
        return {}

    def change_video_settings(self, options):
        pass

    def start_video_stream(self, stream_name, options):
        raise NotImplementedError("You need to write this!")



import logging
import os
import sys
import subprocess
import tempfile

from unifi.cams.base import UnifiCamBase

FNULL = open(os.devnull, "w")


class RTSPCam(UnifiCamBase):
    @classmethod
    def add_parser(self, parser):
        parser.add_argument("--source", "-s", required=True, help="Stream source")
        parser.add_argument(
            "--ffmpeg-args",
            "-f",
            default="-vcodec copy -strict -2 -c:a aac",
            help="Transcoding args for `ffmpeg -i <src> <args> <dst>`",
        )

    def __init__(self, args, logger=None):
        self.logger = logger
        self.args = args
        self.dir = tempfile.mkdtemp()
        self.logger.info(self.dir)
        cmd = 'ffmpeg -y -re -i "{}" -vf fps=1 -update 1 {}/screen.jpg'.format(
            self.args.source, self.dir
        )
        self.logger.info(cmd)
        self.streams = {
            "mjpg": subprocess.Popen(
                cmd, stdout=FNULL, stderr=subprocess.STDOUT, shell=True
            )
        }

    def get_snapshot(self):
        return "{}/screen.jpg".format(self.dir)

    def start_video_stream(self, stream_name, options):
        cmd = 'ffmpeg -y -f lavfi -i aevalsrc=0 -i "{}" {} -metadata streamname={} -f flv - | {} -m unifi.clock_sync | nc {} 6666'.format(
            self.args.source,
            self.args.ffmpeg_args,
            stream_name,
            sys.executable,
            self.args.host,
        )
        self.logger.info("Spawning ffmpeg (%s): %s", stream_name, cmd)
        if (
            stream_name not in self.streams
            or self.streams[stream_name].poll() is not None
        ):
            self.streams[stream_name] = subprocess.Popen(
                cmd, stdout=FNULL, stderr=subprocess.STDOUT, shell=True
            )
